#include <LiquidCrystal_I2C.h>
#include <string.h>

#define N 4
#define GET(s, i, j) (s)[N * (i) + (j)]

#define BUZZER 3
#define NOTE_LEN 175

#define LEFT 2
#define DOWN 4
#define UP 12
#define RIGHT 13

#define CLEAR 6
#define TOGGLE 7
#define CHECK 8

#define BLUE 5
#define GREEN 9
#define RED 10

LiquidCrystal_I2C lcd(0x27, 20, 4);

char boards[70][16] = {

	// The EASY boards

	"21___32____41___",
	"1_34_3__2______1",
	"__12_1_3____4_2_",
	"_23__3_____12_4_",
	"42__1____12____3",
	"___34____13_3_2_",
	"_32____41_____12",
	"4_____1__24___21",
	"_2____2_3_4_2__1",
	"___14__22__4__23",
	"4______22_14_4__",
	"_21___32_1__2__1",
	"4_21___3_____23_",
	"24_3_____34____1",
	"4____34__43__1_4",
	"__122______443__",
	"4_____4__12_2__3",
	"3______2_1_3__21",
	"___2_2___3_4_12_",
	"3__11____123____",
	"__3_23__1___3__4",
	"_3__1_2___4_4__1",
	"___112___4___13_",
	"__3__1_44__31_4_",
	"___343_2_12_____",
	"_3__2____2_14__2",
	"1_4___12_1__4__1",
	"__4_1_3__1_4___3",
	"__1__2__4_2_2__4",
	"_3__4___123_3___",
	"___3_4_2__211___",
	"_2_1_43____44___",
	"___1_4__4312____",
	"2__1__4__1___2_3",
	"43_1_2___1_3__1_",


	// The HARD boards

	"3__11_________2_",
	"4_____4____3_2__",
	"__1_3____3_____4",
	"4__3_2__2_4_____",
	"__1_3____3_____4",
	"4__3_2__2_4_____",
	"_2____2____34___",
	"2______3_2____1_",
	"__4_1____2_____3",
	"2______3_____41_",
	"_4_____2______23",
	"_4__3_____14__2_",
	"___3___43___2___",
	"1______1_4____2_",
	"__2_1____4_____3",
	"14____________24",
	"____43_22__4____",
	"_1____2____34___",
	"___44_____1__2__",
	"__1_3____3_____2",
	"_2_1__2_41______",
	"___32____4____1_",
	"1_2_____3__4____",
	"_1_____4__3_2___",
	"_____2____1__3_2",
	"_2_1____2_3_____",
	"_1_____1___2_4__",
	"2___1_2____4____",
	"_____2___43___1_",
	"___23_____1__1__",
	"3____4___1_____2",
	"_____1____2_2_3_",
	"___2____4____2_1",
	"__2__3_____24___",
	"___2_____41_3___"
};

char original[16] = {0};
char board[16] = {0};
char prompt[20];

char X, Y, lives, aux;

byte CHAR_HEART[8] = {0, 10, 31, 31, 14, 4, 0, 0};
byte CHAR_SMILE[8] = {0, 10, 10, 0, 31, 31, 14, 0};
byte CHAR_HAPPY[8] = {0, 10, 21, 0, 31, 31, 14, 0};
byte CHAR_SAD[8] = {0, 10, 10, 0, 14, 17, 0, 0};

bool chooseDifficulty;

/**
 * Plays "We Are Number One" by Stephan Karl Stephenson
 * as the intro for the game.
 */
void playIntro() {
	tone(BUZZER, 440);
	delay(NOTE_LEN * 2);
	noTone(BUZZER);
	delay(NOTE_LEN);
	tone(BUZZER, 659);
	delay(NOTE_LEN);

	tone(BUZZER, 622);
	delay(NOTE_LEN / 2);
	tone(BUZZER, 659);
	delay(NOTE_LEN / 2);
	tone(BUZZER, 622);
	delay(NOTE_LEN / 2);
	tone(BUZZER, 659);
	delay(NOTE_LEN / 2);
	tone(BUZZER, 622);
	delay(NOTE_LEN);
	tone(BUZZER, 659);
	delay(NOTE_LEN);

	tone(BUZZER, 523);
	delay(NOTE_LEN * 2);
	tone(BUZZER, 440);
	delay(NOTE_LEN * 2);

	noTone(BUZZER);
	delay(NOTE_LEN);
	tone(BUZZER, 440);
	delay(NOTE_LEN);
	tone(BUZZER, 523);
	delay(NOTE_LEN);
	tone(BUZZER, 659);
	delay(NOTE_LEN);

	tone(BUZZER, 698);
	delay(NOTE_LEN * 2);
	tone(BUZZER, 523);
	delay(NOTE_LEN * 2);

	tone(BUZZER, 698);
	delay(NOTE_LEN * 2);
	tone(BUZZER, 783);
	delay(NOTE_LEN * 2);

	tone(BUZZER, 659);
	delay(NOTE_LEN);
	tone(BUZZER, 698);
	delay(NOTE_LEN);
	tone(BUZZER, 659);
	delay(NOTE_LEN);
	tone(BUZZER, 698);
	delay(NOTE_LEN);

	tone(BUZZER, 659);
	delay(NOTE_LEN);
	noTone(BUZZER);
	delay(NOTE_LEN);
	tone(BUZZER, 1318);
	delay(NOTE_LEN);
	noTone(BUZZER);
	delay(NOTE_LEN);

	tone(BUZZER, 440);
	delay(NOTE_LEN);
	noTone(BUZZER);
	delay(NOTE_LEN);
}

/**
 * Plays a short song to celebrate the victory.
 */
void playVictoryJingle() {
	analogWrite(GREEN, 255);
	tone(BUZZER, 659);
	delay(250);
	tone(BUZZER, 523);
	delay(250);
	tone(BUZZER, 784);
	delay(250);
	tone(BUZZER, 1047);
	delay(1250);
	noTone(BUZZER);
	analogWrite(GREEN, 0);
}

/**
 * Plays a short song to make you feel bad about your decisions.
 */
void playGameOverJingle() {
	analogWrite(RED, 255);
	tone(BUZZER, 880);
	delay(250);
	tone(BUZZER, 659);
	delay(250);
	tone(BUZZER, 622);
	delay(250);
	tone(BUZZER, 523);
	delay(250);
	tone(BUZZER, 440);
	delay(1000);
	noTone(BUZZER);
	analogWrite(RED, 0);
}

/**
 * Plays a short sound of a given note.
 */
void beep(short note) {
	tone(BUZZER, note);
	delay(100);
	noTone(BUZZER);
	delay(100);
}

/**
 * Prints the game board on the left side of the screen,
 * while also printing the remaining lives, the cursor coodrinates
 * and the character at the current position.
 */
void printGameScreen() {
	// print the game board
	lcd.noCursor();
	for (char i = 0; i < 4; ++i)
	for (char j = 0; j < 4; ++j) {
		lcd.setCursor((j > 1) ? j+1 : j, i);
		lcd.print(board[4 * i + j]);
	}

	// print the board-prompt separator
	for (char i = 0; i < 4; ++i) {
		lcd.setCursor(6, i);
		lcd.write('|');
	}

	// print coordinates
	sprintf(prompt, "L%d, C%d : %c", Y + 1, X + 1, GET(board, Y, X));
	lcd.setCursor(8, 0);
	lcd.print(prompt);

	// print lives
	lcd.setCursor(15, 1);
	lcd.write((lives == 3) ? (byte)0 : ' ');
	lcd.write((lives >= 2) ? (byte)0 : ' ');
	lcd.write((lives >= 1) ? (byte)0 : ' ');

	// set the cursor position to the (X, Y) coordinates
	lcd.setCursor((X >= 2) ? X + 1 : X, Y);
	lcd.cursor();
	delay(5);
}

/**
 * Prints a warning message,
 * plays an adequate tone to warn the player
 * and writes on the screen the coresponding message.
 */
void printWarning() {
	analogWrite(RED, 255);
	analogWrite(BLUE, 255);
	lcd.setCursor(8, 2);
	lcd.print("Can't change");
	lcd.setCursor(8, 3);
	lcd.print("at this spot");
	tone(BUZZER, 740);
	delay(500);

	analogWrite(RED, 0);
	analogWrite(BLUE, 0);
	lcd.setCursor(8, 3);
	lcd.print("            ");
	lcd.setCursor(8, 2);
	lcd.print("            ");
	noTone(BUZZER);
}

/**
 * The Gamve Over screen and prompt
 */
void gameOver() {
	digitalWrite(RED, HIGH);
	lcd.clear();
	lcd.noCursor();
	lcd.setCursor(3, 0);
	lcd.print("GAME OVER ");
	lcd.write((byte)2);
	lcd.setCursor(4, 2);
	lcd.print("Better luck");
	lcd.setCursor(4, 3);
	lcd.print("next time.");
	playGameOverJingle();
	digitalWrite(RED, LOW);

	lcd.setCursor(2, 0);
	lcd.print("Press RESET on the");
	lcd.setCursor(0, 1);
	lcd.print("Arduino board if you");
	lcd.setCursor(0, 2);
	lcd.print("want to play another");
	lcd.setCursor(0, 3);
	lcd.print("round of Sudoku 4x4.");

	exit(0);
}

/**
 * Prints a good message if the player has checked his progress
 * and he has not made any mistakes in his game.
 */
void printGoodCheck(char no_spaces) {
	lcd.noCursor();
	analogWrite(BLUE, 255);
	analogWrite(GREEN, (no_spaces <= getFreeSpaces(original) / 2) ? 255 : 0);
	lcd.setCursor(8, 2);
	lcd.print("So far,");
	lcd.setCursor(8, 3);
	lcd.print("so good. ");
  lcd.write((byte)1);
	tone(BUZZER, 1047);
	delay(500);

	analogWrite(BLUE, 0);
	analogWrite(GREEN, 0);

	lcd.setCursor(8, 2);
	sprintf(prompt, "%d spaces", no_spaces);
	lcd.print(prompt);
	lcd.setCursor(8, 3);
	lcd.print("left      ");
	noTone(BUZZER);
	delay(500);

	lcd.setCursor(8, 2);
	lcd.print("         ");
	lcd.setCursor(8, 3);
	lcd.print("          ");      
}

/**
 * Prints a message which congratulates the player for completing
 * the current stage.
 */
void printLevelComplete() {
	lcd.clear();
	lcd.noCursor();

	analogWrite(GREEN, 255);
	lcd.setCursor(2, 1);
	lcd.print("LEVEL COMPLETE!");
	lcd.setCursor(2, 2);
	lcd.print("CONGRATULATIONS!");
	playVictoryJingle();
	analogWrite(GREEN, 0);

	lcd.clear();
	lcd.setCursor(0, 0);
	lcd.print("Thank you for");
	lcd.setCursor(0, 1);
	lcd.print("playing my game!");
	lcd.setCursor(0, 2);
	lcd.print("I hope you enjoyed");
	lcd.setCursor(0, 3);
	lcd.print("playing it!        ");
	lcd.write((byte)3);

	exit(0);
}

/**
 * Prints a message to tell the player that he has got at least 1 mistake
 * on his board at the last check
 */
void printBadCheck(char lives) {
	analogWrite(RED, 85 * (3 - lives));
	lcd.noCursor();
	lcd.setCursor(8, 2);
	lcd.print("Bad value on");
	lcd.setCursor(8, 3);
	lcd.print("this board");
	tone(BUZZER, 622);
	delay(500);

	analogWrite(RED, 0);
	lcd.setCursor(8, 2);
	sprintf(prompt, "%d lives left", lives);
	lcd.print(prompt);
	lcd.setCursor(8, 3);
	lcd.print("          ");
	noTone(BUZZER);
	delay(500);

	lcd.setCursor(8, 2);
	lcd.print("            ");
}

/**
 * Returns "true" if the given board element is placed correctly compared to
 * the elements from the same line, from the same column and from the same
 * coresponding 2x2 block.
 * Retruns "false", otherwise.
 */
bool checkBoardElement(char board[16], char original[16], char y, char x) {
	char i, j;
	char val = GET(board, y, x);

	// Same row equality
	for (j = 0; j < N; ++j) {
		if (j == x || GET(board, y, j) == GET(original, y, j))
			continue;

		if (GET(board, y, j) == val)
			return false;
	}

	// Same column equality
	for (i = 0; i < N; ++i) {
		if (i == y || GET(board, i, x) == GET(original, i, x))
			continue;

		if (GET(board, i, x) == val)
			return false;
	}

	// Same block equality
	char yB = (y / 2) * 2;
	char xB = (x / 2) * 2;

	for (i = yB; i <= yB + 1; ++i)
		for (j = xB; j <= xB + 1; ++j) {
			if ((i == y && j == x) || GET(board, i, j) == GET(original, i, j))
				continue;
			if (GET(board, i, j) == val)
				return false;
		}
  
	return true;
}

/**
 * Returns "false" if we have 1 misplaced number;
 * Returns "true", otherwise. 
 */
bool checkBoard(char board[16], char original[16]) {
	char i, j;
	for (i = 0; i < N; ++i)
		for (j = 0; j < N; ++j) {
			if (GET(board, i,  j) == ' ')
				continue;
			if (!checkBoardElement(board, original, i, j))
				return false;
		}
	return true;
}

/**
 * Returns the number of empty spaces from the given board.
 */
char getFreeSpaces(char board[16]) {
	char spaces = 0, i, j;
	for (i = 0; i < N; ++i)
		for (j = 0; j < N; ++j)
			if (GET(board, j, i) == '_')
				++spaces;
	return spaces;
}

/**
 * Initializes the pins, the LCD screen and the necessary variables.
 */
void setup() {
	// Start the LCD
	lcd.init();
	lcd.backlight();
	lcd.createChar(0, CHAR_HEART);
	lcd.createChar(1, CHAR_SMILE);
	lcd.createChar(2, CHAR_SAD);
	lcd.createChar(3, CHAR_HAPPY);

	lives = 3;
	randomSeed(analogRead(0));

	// Initalize the pins
	pinMode(LEFT, INPUT);
	pinMode(DOWN, INPUT);
	pinMode(UP, INPUT);
	pinMode(RIGHT, INPUT);
	pinMode(CLEAR, INPUT);
	pinMode(TOGGLE, INPUT);
	pinMode(CHECK, INPUT);

	pinMode(BUZZER, OUTPUT);
	pinMode(BLUE, OUTPUT);
	pinMode(GREEN, OUTPUT);
	pinMode(RED, OUTPUT);

	// First screen text
	lcd.noCursor();
	lcd.setCursor(1, 0);
	lcd.print("Sudoku 4x4 on LCD");

	lcd.setCursor(0, 1);
	lcd.print("--------------------");

	lcd.setCursor(1, 2);
	lcd.print("Bogdan-Andrei Buga");

	lcd.setCursor(3, 3);
	lcd.print("PM 2021, 332CB");

	playIntro();
	lcd.clear();

	// Go to the next state
	chooseDifficulty = true;
}

/**
 * The game playthrough.
 */
void loop() {
/*----------------------------------------------------------------------------
 Difficulty Choice
----------------------------------------------------------------------------*/
	if (chooseDifficulty) {

		lcd.setCursor(1, 0);
		lcd.print("Choose difficulty");
		lcd.setCursor(0, 1);
		lcd.print("--------------------");
		lcd.setCursor(1, 2);
		lcd.print("Press UP for EASY");
		lcd.setCursor(1, 3);
		lcd.print("Press DOWN for HARD");

		// Have we pressed either UP or DOWN?
		if (digitalRead(UP) == HIGH || digitalRead(DOWN) == HIGH) {
			char idx = random(35);

			// Get the board from the boards array.
			// Go on HARD mode only if we pressed DOWN, without pressing UP.
			if (digitalRead(UP) == LOW && digitalRead(DOWN) == HIGH)
			idx += 35;

			// Initialize the board and the coordinates
			strncpy(original, boards[idx], 16);
			strncpy(board, original, 16);
			X = 0;
			Y = 0;

			// Exit the current state and play the game
			lcd.clear();
			chooseDifficulty = false;
			delay(100);
			return;
		}
		return;
	}

/*----------------------------------------------------------------------------
 Player - Button interactions
----------------------------------------------------------------------------*/
	// UP
	if (digitalRead(UP) == HIGH) {
		Y = max(Y - 1, 0);
		beep(880);
	}

	// DOWN
	if (digitalRead(DOWN) == HIGH) {
		Y = min(Y + 1, N - 1);
		beep(880);
	}

	// LEFT
	if (digitalRead(LEFT) == HIGH) {
		X = max(X - 1, 0);
		beep(880);
	}

	// RIGHT
	if (digitalRead(RIGHT) == HIGH) {
		X = min(X + 1, N - 1);
		beep(880);
	}

	// CLEAR / TOGGLE
	if (digitalRead(TOGGLE) == HIGH || digitalRead(CLEAR) == HIGH) {

		// Unchangable spot
		if (GET(original, Y, X) != '_') {
			printWarning();
			return;
		}

		// Change the current value
		aux = GET(board, Y, X);
		if (digitalRead(TOGGLE) == HIGH)
			aux = (aux == '_') ? '1' : ((aux == '4') ? '1' : aux + 1);
		else if (digitalRead(CLEAR) == HIGH)
			aux = '_';
		GET(board, Y, X) = aux;
		beep(659);
	}

	// CHECK
	if (digitalRead(CHECK) == HIGH) {

		// Is there any mistake on our current board?
		bool check = checkBoard(board, original);
		if (!check) {
			if (!(--lives))
			gameOver();

			printBadCheck(lives);

		} else {
			char no_spaces = getFreeSpaces(board);
			if (!no_spaces)
				printLevelComplete();
			else
				printGoodCheck(no_spaces);
		}
	}
  
/*----------------------------------------------------------------------------
 Print the results on the LCD Screen
----------------------------------------------------------------------------*/
	printGameScreen();
}
